import logging
from pathlib import Path

import click
import ruamel.yaml
from tabulate import tabulate

from trailblazer.mip import files

LOG = logging.getLogger(__name__)


@click.command()
@click.argument('family')
@click.pass_context
def check(context: click.Context, family: str):
    """Delete an analysis log from the database."""
    analysis_obj = context.obj['store'].analyses(family=family).first()
    if analysis_obj is None:
        LOG.error('no analysis found')
        context.abort()

    config_path = Path(analysis_obj.config_path)
    if not config_path.exists():
        LOG.error(f"analysis config not found: {config_path}")
        context.abort()
    config_raw = ruamel.yaml.safe_load(config_path.open())
    config_data = files.parse_config(config_raw)

    sampleinfo_raw = ruamel.yaml.safe_load(Path(config_data['sampleinfo_path']).open())
    sampleinfo_data = files.parse_sampleinfo(sampleinfo_raw)

    qcmetrics_path = Path(sampleinfo_data['qcmetrics_path'])
    if not qcmetrics_path.exists():
        LOG.error(f"qc metrics not found: {str(qcmetrics_path)}")
        context.abort()
    qcmetrics_raw = ruamel.yaml.safe_load(qcmetrics_path.open())
    qcmetrics_data = files.parse_qcmetrics(qcmetrics_raw)

    samples = {
        'sample': [],
        'type': [],
        'ped': [],
        'chanjo': [],
        'peddy': [],
        'plink': [],
        'duplicates': [],
    }
    for sample_data in config_data['samples']:
        LOG.debug(f"{sample_data['id']}: parse analysis config")
        samples['sample'].append(sample_data['id'])
        samples['type'].append(sample_data['type'])

    for sample_data in sampleinfo_data['samples']:
        LOG.debug(f"{sample_data['id']}: parse sample info")
        samples['ped'].append(sample_data['sex'])

        with Path(sample_data['chanjo_sexcheck']).open() as chanjo_handle:
            sexcheck_data = files.parse_chanjo_sexcheck(chanjo_handle)

        predicted_sex = sexcheck_data['predicted_sex']
        xy_ratio = sexcheck_data['y_coverage'] / sexcheck_data['x_coverage']
        samples['chanjo'].append(f"{predicted_sex} ({xy_ratio:.3f})")

    for sample_data in qcmetrics_data['samples']:
        LOG.debug(f"{sample_data['id']}: parse qc metrics")
        samples['plink'].append(sample_data['plink_sex'])
        duplicates_percent = sample_data['duplicates'] * 100
        samples['duplicates'].append(f"{duplicates_percent:.3f}%")

    peddy_path = Path(sampleinfo_data['peddy']['sex_check'])
    if peddy_path.exists():
        with peddy_path.open() as sexcheck_handle:
            peddy_data = files.parse_peddy_sexcheck(sexcheck_handle)

        for sample_id in samples['sample']:
            LOG.debug(f"{sample_id}: parse peddy")
            predicted_sex = peddy_data[sample_id]['predicted_sex']
            het_ratio = peddy_data[sample_id]['het_ratio']
            samples['peddy'].append(f"{predicted_sex} ({het_ratio})")
    else:
        LOG.warning(f"missing peddy output: {peddy_path}")

    print(tabulate(samples, headers='keys', tablefmt='psql'))



# -*- coding: utf-8 -*-
import os


def environ_email():
    """Guess email from sudo user environment variable."""
    username = os.environ.get('SUDO_USER')
    if username:
        return f"{username}@scilifelab.se"



# -*- coding: utf-8 -*-
import datetime

import alchy
from sqlalchemy import Column, ForeignKey, orm, types, UniqueConstraint

from trailblazer.mip import sacct
from trailblazer.constants import TEMP_STATUSES

STATUS_OPTIONS = ('pending', 'running', 'completed', 'failed', 'error', 'canceled')
JOB_STATUS_OPTIONS = [category.lower() for category in sacct.CATEGORIES]
PRIORITY_OPTIONS = ('low', 'normal', 'high')
TYPES = ('wes', 'wgs', 'rna')

Model = alchy.make_declarative_base(Base=alchy.ModelBase)


class Info(Model):

    """Keep track of meta data."""

    __tablename__ = 'info'

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.datetime.now)
    updated_at = Column(types.DateTime)


class User(Model):

    __tablename__ = 'user'

    id = Column(types.Integer, primary_key=True)
    google_id = Column(types.String(128), unique=True)
    email = Column(types.String(128), unique=True)
    name = Column(types.String(128))
    avatar = Column(types.Text)
    created_at = Column(types.DateTime, default=datetime.datetime.now)

    runs = orm.relationship('Analysis', backref='user')

    @property
    def first_name(self) -> str:
        """First part of name."""
        return self.name.split(' ')[0]


class Analysis(Model):

    """Analysis record."""

    __tablename__ = 'analysis'
    __table_args__ = (UniqueConstraint('family', 'started_at', 'status',
                                       name='_uc_family_start_status'),)

    id = Column(types.Integer, primary_key=True)
    family = Column(types.String(128), nullable=False)

    version = Column(types.String(32))
    logged_at = Column(types.DateTime, default=datetime.datetime.now)
    started_at = Column(types.DateTime)
    completed_at = Column(types.DateTime)
    status = Column(types.Enum(*STATUS_OPTIONS))
    priority = Column(types.Enum(*PRIORITY_OPTIONS))
    out_dir = Column(types.Text)
    config_path = Column(types.Text)
    comment = Column(types.Text)
    is_deleted = Column(types.Boolean, default=False)
    is_visible = Column(types.Boolean, default=True)
    type = Column(types.Enum(*TYPES))
    user_id = Column(ForeignKey(User.id))
    progress = Column(types.Float, default=0.)

    failed_jobs = orm.relationship('Job', backref='analysis')

    @property
    def is_temp(self):
        """Check if the log is for a temporary status: running/pending."""
        return self.status in TEMP_STATUSES


class Job(Model):

    """Represent a step in the pipeline."""

    __tablename__ = 'job'

    id = Column(types.Integer, primary_key=True)
    analysis_id = Column(ForeignKey(Analysis.id, ondelete='CASCADE'), nullable=False)
    slurm_id = Column(types.Integer)
    name = Column(types.String(64))
    context = Column(types.String(64))
    started_at = Column(types.DateTime)
    elapsed = Column(types.Integer)
    status = Column(types.Enum(*JOB_STATUS_OPTIONS))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Trailblazer is a tool to manage and track state of analyses.

You can install it from PyPI as follows::
    pip install trailblazer

For more detailed instructions, run :code:`trailblazer --help`

See the GitHub repo for documentation:
`Clinical-Genomics/analysis <https://github.com/Clinical-Genomics/analysis>`_
"""
# To use a consistent encoding
import codecs
import os
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

# Shortcut for building/publishing to Pypi
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel upload')
    sys.exit()


def parse_reqs(req_path='./requirements.txt'):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with codecs.open(req_path, 'r') as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle
                 if line.strip() and not line.startswith('#'))
        for line in lines:
            # check for nested requirements files
            if line.startswith('-r'):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])
            else:
                # add the line as a new requirement
                install_requires.append(line)
    return install_requires


# This is a plug-in for setuptools that will invoke py.test
# when you run python setup.py test
class PyTest(TestCommand):

    """Set up the py.test test runner."""

    def finalize_options(self):
        """Set options for the command line."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Execute the test runner command."""
        # Import here, because outside the required eggs aren't loaded yet
        import pytest
        sys.exit(pytest.main(self.test_args))


setup(
    name='trailblazer',

    # Versions should comply with PEP440. For a discussion on
    # single-sourcing the version across setup.py and the project code,
    # see http://packaging.python.org/en/latest/tutorial.html#version
    version='4.0.1',

    description=('Track MIP analyses.'),
    long_description=__doc__,
    # What does your project relate to? Separate with spaces.
    keywords='analysis development',
    author='Robin Andeer',
    author_email='robin.andeer@scilifelab.se',
    license='MIT',

    # The project's main homepage
    url='https://github.com/Clinical-Genomics/analysis',

    packages=find_packages(exclude=('tests*', 'docs', 'examples')),

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    include_package_data=True,
    # package_data={
    #     'taboo': [
    #         'server/genotype/templates/genotype/*.html',
    #     ]
    # },
    zip_safe=False,

    # Although 'package_data' is the preferred approach, in some case you
    # may need to place data files outside of your packages.
    # In this case, 'data_file' will be installed into:
    # '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # Install requirements loaded from ``requirements.txt``
    install_requires=parse_reqs(),
    tests_require=[
        'pytest',
    ],
    cmdclass=dict(
        test=PyTest,
    ),

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and
    # allow pip to create the appropriate form of executable for the
    # target platform.
    entry_points={
        'console_scripts': [
            'trailblazer = trailblazer.cli:base',
        ]
    },

    # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',

        'Environment :: Console',
    ],
)



# -*- coding: utf-8 -*-
""" Get MIP data for trending."""

from trailblazer.mip import files


def parse_mip_analysis(mip_config_raw: dict, qcmetrics_raw: dict, sampleinfo_raw: dict) -> dict:
    """Parse the output analysis files from MIP for adding info
    to trend database

    Args:
        mip_config_raw (dict): raw YAML input from MIP analysis config file
        qcmetrics_raw (dict): raw YAML input from MIP analysis qc metric file
        sampleinfo_raw (dict): raw YAML input from MIP analysis qc sample info file
    Returns:
        dict: parsed data
    """
    outdata = _define_output_dict()

    _config(mip_config_raw, outdata)
    _qc_metrics(outdata, qcmetrics_raw)
    _qc_sample_info(outdata, sampleinfo_raw)

    return outdata


def _add_rank_model_version(outdata, sampleinfo_data):
    outdata['sv_rank_model_version'] = sampleinfo_data['sv_rank_model_version']
    outdata['rank_model_version'] = sampleinfo_data['rank_model_version']


def _qc_sample_info(outdata, sampleinfo_raw):
    sampleinfo_data = _parse_qc_sample_info_file(sampleinfo_raw)
    _add_mip_version(outdata, sampleinfo_data)
    _add_genome_build(outdata, sampleinfo_data)
    _add_rank_model_version(outdata, sampleinfo_data)


def _qc_metrics(outdata, qcmetrics_raw):
    qcmetrics_data = _parse_qc_metric_file_into_dict(qcmetrics_raw)
    _add_sample_level_info_from_qc_metric_file(outdata, qcmetrics_data)


def _config(mip_config_raw, outdata):
    config_data = _parse_raw_mip_config_into_dict(mip_config_raw)
    _add_case_id(config_data, outdata)
    _add_all_samples_from_mip_config(config_data, outdata)


def _add_genome_build(outdata, sampleinfo_data):
    outdata['genome_build'] = sampleinfo_data['genome_build']


def _add_mip_version(outdata, sampleinfo_data):
    outdata['mip_version'] = sampleinfo_data['version']


def _parse_qc_sample_info_file(sampleinfo_raw):
    sampleinfo_data = files.parse_sampleinfo(sampleinfo_raw)
    return sampleinfo_data


def _add_sample_level_info_from_qc_metric_file(outdata, qcmetrics_data):
    for sample_data in qcmetrics_data['samples']:
        _add_duplicate_reads(outdata, sample_data)
        _add_mapped_reads(outdata, sample_data)
        _add_predicted_sex(outdata, sample_data)
        _add_dropout_rates(outdata, sample_data)
        _add_insert_size_metrics(outdata, sample_data)


def _add_dropout_rates(outdata, sample_data):
    outdata['at_dropout'][sample_data['id']] = sample_data['at_dropout']
    outdata['gc_dropout'][sample_data['id']] = sample_data['gc_dropout']


def _add_insert_size_metrics(outdata, sample_data):
    outdata['median_insert_size'][sample_data['id']] = sample_data['median_insert_size']
    outdata['insert_size_standard_deviation'][sample_data['id']] = sample_data['insert_size_standard_deviation']


def _add_predicted_sex(outdata, sample_data):
    outdata['analysis_sex'][sample_data['id']] = sample_data['predicted_sex']


def _add_mapped_reads(outdata, sample_data):
    mapped_reads_percent = sample_data['mapped'] * 100
    outdata['mapped_reads'][sample_data['id']] = mapped_reads_percent


def _add_duplicate_reads(outdata, sample_data):
    duplicates_percent = sample_data['duplicates'] * 100
    outdata['duplicates'][sample_data['id']] = duplicates_percent


def _parse_qc_metric_file_into_dict(qcmetrics_raw):
    qcmetrics_data = files.parse_qcmetrics(qcmetrics_raw)
    return qcmetrics_data


def _parse_raw_mip_config_into_dict(mip_config_raw):
    config_data = files.parse_config(mip_config_raw)
    return config_data


def _define_output_dict():
    outdata = {
        'analysis_sex': {},
        'at_dropout': {},
        'case': None,
        'duplicates': {},
        'gc_dropout': {},
        'genome_build': None,
        'rank_model_version': None,
        'insert_size_standard_deviation': {},
        'mapped_reads': {},
        'median_insert_size': {},
        'mip_version': None,
        'sample_ids': [],
    }

    return outdata


def _add_all_samples_from_mip_config(config_data, outdata):
    for sample_data in config_data['samples']:
        outdata['sample_ids'].append(sample_data['id'])


def _add_case_id(config_data, outdata):
    outdata['case'] = config_data['case']



# -*- coding: utf-8 -*-
import logging
import signal
import subprocess

from trailblazer.exc import MipStartError

LOG = logging.getLogger(__name__)

CLI_OPTIONS = {
    'config': {'option': '--config_file'},
    'priority': {'option': '--slurm_quality_of_service'},
    'email': {'option': '--email'},
    'base': {'option': '--cluster_constant_path'},
    'dryrun': {'option': '--dry_run_all'},
    'gene_list': {'option': '--vcfparser_slt_fl'},
    'max_gaussian': {'option': '--gatk_varrecal_snv_max_gau'},
    'skip_evaluation': {'option': '--qccollect_skip_evaluation'},
    'start_with': {'option': '--start_with_recipe'},
}


class MipCli(object):

    """Wrapper around MIP command line interface."""

    def __init__(self, script, pipeline):
        """Initialize MIP command line interface."""
        self.script = script
        self.pipeline = pipeline

    def __call__(self, config, case, **kwargs):
        """Execute the pipeline."""
        command = self.build_command(case, config, **kwargs)
        LOG.debug(' '.join(command))
        process = self.execute(command)
        process.wait()
        if process.returncode != 0:
            raise MipStartError('error starting analysis, check the output')
        return process

    def build_command(self, case, config, **kwargs):
        """Builds the command to execute MIP."""
        command = [self.script, self.pipeline, case, CLI_OPTIONS['config']['option'], config]
        for key, value in kwargs.items():
            # enable passing in flags as "False" - shouldn't add command
            if value:
                command.append(CLI_OPTIONS[key]['option'])

                # append value for non-flags
                if value is not True:
                    command.append(value)
        return command

    def execute(self, command):
        """Start a new MIP run."""

        """
        Remove the default SIGPIPE handler
        https://blog.nelhage.com/2010/02/a-very-subtle-bug/
        """
        process = subprocess.Popen(
            command,
            preexec_fn=lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        )
        return process
